﻿// *********************************************************
//
// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in
// compliance with the License. You may obtain a copy of
// the License at
//
// http://www.apache.org/licenses/LICENSE-2.0 
//
// THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES
// OR CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED,
// INCLUDING WITHOUT LIMITATION ANY IMPLIED WARRANTIES
// OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR
// PURPOSE, MERCHANTABILITY OR NON-INFRINGEMENT.
//
// See the Apache 2 License for the specific language
// governing permissions and limitations under the License.
//
// *********************************************************

using System;
using Roslyn.Compilers;
using Roslyn.Compilers.Common;
using Roslyn.Samples.SyntaxVisualizer.Control;
using CSharp = Roslyn.Compilers.CSharp;
using VisualBasic = Roslyn.Compilers.VisualBasic;

namespace Roslyn.Samples.SyntaxVisualizer.Debugger
{
    // A transporter that represents a serialized SyntaxTree / SyntaxNode / SyntaxToken /
    // SyntaxTrivia. This transporter serializes the FullText of the root SyntaxNode of the
    // SyntaxTree being debugged / of the SyntaxTree containing the SyntaxNode / SyntaxToken /
    // SyntaxTrivia being debugged to the debugger side. On the debugger side this FullText is
    // re-parsed to get a SyntaxTree that can be displayed visually.
    [Serializable]
    internal class SyntaxTransporter
    {
        #region Serialized Properties
        // The language (C# / VB) for the SyntaxTree / SyntaxNode / SyntaxToken /
        // SyntaxTrivia being serialized.
        internal string SourceLanguage { get; private set; }

        // The FullText for the root SyntaxNode of the SyntaxTree being serialized / of the
        // SyntaxTree containing the SyntaxNode / SyntaxToken / SyntaxTrivia being serialized.
        internal string SyntaxTreeSourceText { get; private set; }

        // The Span for the root SyntaxNode of the SyntaxTree being serialized / of the
        // SyntaxNode / SyntaxToken / SyntaxTrivia being serialized.
        internal TextSpan ItemSpan { get; private set; }

        // The SyntaxKind for the root SyntaxNode of the SyntaxTree being serialized / of the
        // SyntaxNode / SyntaxToken / SyntaxTrivia being serialized.
        internal string ItemKind { get; private set; }
        #endregion

        #region Helpers - GetRootNode
        // Helpers that return the root SyntaxNode for a supplied SyntaxNode /
        // SyntaxToken / SyntaxTrivia.

        private CommonSyntaxNode GetRootNode(CommonSyntaxNode node)
        {
            while (node != null && node.Parent != null)
            {
                node = node.Parent;
            }

            return node;
        }

        private CommonSyntaxNode GetRootNode(CommonSyntaxNodeOrToken nodeOrToken)
        {
            if (nodeOrToken.IsNode)
            {
                return GetRootNode(nodeOrToken.AsNode());
            }
            else
            {
                return GetRootNode(nodeOrToken.AsToken());
            }
        }

        private CommonSyntaxNode GetRootNode(CommonSyntaxToken token)
        {
            return GetRootNode(token.Parent);
        }

        private CommonSyntaxNode GetRootNode(CommonSyntaxTrivia trivia)
        {
            return GetRootNode(trivia.Token);
        }
        #endregion

        #region Constructors
        internal SyntaxTransporter(CommonSyntaxTree tree)
        {
            if (tree is CSharp.SyntaxTree)
            {
                SourceLanguage = LanguageNames.CSharp;
            }
            else
            {
                SourceLanguage = LanguageNames.VisualBasic;
            }

            SyntaxTreeSourceText = tree.Root.GetFullText();
            ItemSpan = tree.Root.Span;
            ItemKind = tree.Root.GetKind(SourceLanguage);
        }

        internal SyntaxTransporter(CommonSyntaxNodeOrToken nodeOrToken)
        {
            try
            {
                var t = (CSharp.SyntaxNodeOrToken)nodeOrToken;
                SourceLanguage = LanguageNames.CSharp;
            }
            catch
            {
                var t = (VisualBasic.SyntaxNodeOrToken)nodeOrToken;
                SourceLanguage = LanguageNames.VisualBasic;
            }

            var rootNode = GetRootNode(nodeOrToken);
            if (rootNode != null)
            {
                SyntaxTreeSourceText = rootNode.GetFullText();
            }

            ItemSpan = nodeOrToken.Span;
            ItemKind = nodeOrToken.GetKind(SourceLanguage);
        }

        internal SyntaxTransporter(CommonSyntaxNode node)
        {
            if (node is CSharp.SyntaxNode)
            {
                SourceLanguage = LanguageNames.CSharp;
            }
            else
            {
                SourceLanguage = LanguageNames.VisualBasic;
            }

            var rootNode = GetRootNode(node);
            if (rootNode != null)
            {
                SyntaxTreeSourceText = rootNode.GetFullText();
            }

            ItemSpan = node.Span;
            ItemKind = node.GetKind(SourceLanguage);
        }

        internal SyntaxTransporter(CommonSyntaxToken token)
        {
            try
            {
                var t = (CSharp.SyntaxToken)token;
                SourceLanguage = LanguageNames.CSharp;
            }
            catch
            {
                var t = (VisualBasic.SyntaxToken)token;
                SourceLanguage = LanguageNames.VisualBasic;
            }

            var rootNode = GetRootNode(token);
            if (rootNode != null)
            {
                SyntaxTreeSourceText = rootNode.GetFullText();
            }

            ItemSpan = token.Span;
            ItemKind = token.GetKind(SourceLanguage);
        }

        internal SyntaxTransporter(CommonSyntaxTrivia trivia)
        {
            try
            {
                var t = (CSharp.SyntaxTrivia)trivia;
                SourceLanguage = LanguageNames.CSharp;
            }
            catch
            {
                var t = (VisualBasic.SyntaxTrivia)trivia;
                SourceLanguage = LanguageNames.VisualBasic;
            }

            var rootNode = GetRootNode(trivia);
            if (rootNode != null)
            {
                SyntaxTreeSourceText = rootNode.GetFullText();
            }

            ItemSpan = trivia.Span;
            ItemKind = trivia.GetKind(SourceLanguage);
        }
        #endregion

        // Helper that returns the SyntaxTree generated by parsing the serialized source text.
        internal CommonSyntaxTree GetSyntaxTree()
        {
            CommonSyntaxTree tree = null;

            if (SourceLanguage == LanguageNames.CSharp)
            {
                tree = CSharp.SyntaxTree.ParseCompilationUnit(SyntaxTreeSourceText);
            }
            else if (SourceLanguage == LanguageNames.VisualBasic)
            {
                tree = VisualBasic.SyntaxTree.ParseCompilationUnit(SyntaxTreeSourceText);
            }

            return tree;
        }
    }
}
